@@ -1,9 +1,9 @@
 # eg.
 # e=f.find('linktext','Download')
 # e.click()
-# e.send_keys(Keys.COMMAND,'a')
-# e.submit()
-# if space in class name, convert to .
+# e.send(KEYS.COMMAND,KEYS.SHIFT,KEYS.ARROW_LEFT)
+# e.send(KEYS.COMMAND,'a')
+# e.submit(), senddelay(), sendsubmit()
 # f.find_all(By.CLASS_NAME,'d2l-datetime-selector-date-input.d2l-edit')  
 # https://stackoverflow.com/questions/17000703/is-string-matches-supported-in-selenium-webdriver-2
 # https://stackoverflow.com/questions/22436789/xpath-ends-with-does-not-work
@@ -11,11 +11,11 @@
 # f.find_all('xpath','//*[contains(@id,"time")][contains(@title,"Date")]')
 # f.findsel('id','activity--history-range-dropdown',timeout=10).select_by_visible_text('history')
 #                                                              .select_by_index(2)  # 0-based
-# f.snap()
+# f.save_screenshot()                                        | f.snap()
 # f.js_click(e)                                              | f.click(e)
 # f.js_rightclick(e)                                         | f.rclick(e)
-# f.js_setvalue(e,value)                                     |
-# f.js_setvalue_via_typing(e,value)  # essentially send keys | f.send_keys(e,value)
+# f.js_setvalue(e,value)                                     | 
+# f.js_setvalue_via_typing(e,value)  # essentially send keys | f.send(e,value)
 # f.execute_script("arguments[0].click();", e)
 # f.execute_script(f"arguments[0].value='{smscode}';", f.find('id','code'))
 # f.driver is the native webdriver object
@@ -78,7 +78,7 @@ class Firefox:
 
     def __init__(
         self,
-        url = None,
+        url = 'https://bot.sannysoft.com/',
         headless: bool = False,
         wire: bool = False, # Use seleniumwire only when necessary, because it would slow things down a bit (not working if session is True)
         session: bool = False, # Use seleniumrequests for cross-session cookies (overrides wire)
@@ -343,6 +343,11 @@ class Firefox:
                     self.driver.switch_to.window(self.driver.window_handles[-1])
                     self.driver.close()
 
+        # https://akarin.dev/2022/02/15/disable-geckodriver-detection-with-addon/
+        # customized addon, xpi is zip
+        # can be found at about:addons
+        HERE=os.path.dirname(os.path.abspath(__file__))
+        self.driver.install_addon(os.path.join(HERE,'firefox.xpi'), temporary=True)
         if url is not None: self.driver.get(url)
 
     # -------------------------------------------------------- Public methods -------------------------------------------------------- #
@@ -513,6 +518,18 @@ class Firefox:
         except TimeoutException: 
             return ''
 
+    def getresp(self,url:str,regexpat,timeout=300,raw=False,*args,**kwargs):
+        """
+        catch matched url, then request the url with headers, returns json (raw=False) or raw resp (raw=True)
+        """
+        import requests
+        resp=self.getreq(url,regexpat,timeout=timeout,raw=True)
+        resp=requests.get(resp.url,headers=resp.headers,*args,**kwargs)
+        if not raw:
+            return resp.json()
+        else:
+            return resp
+
     def refresh(self,seconds=5) -> None:
         self.driver.refresh()
         time.sleep(seconds)
@@ -528,6 +545,8 @@ class Firefox:
         timeout: Optional[int] = None,
         element: Optional = None
     ) -> Union[Optional[WebElement], List[WebElement]]:
+        # if space in class name, convert to .
+        key=key.strip().replace(' ','.')
         return self.__find(
             by,
             EC.presence_of_element_located,
@@ -575,6 +594,7 @@ class Firefox:
         timeout: Optional[int] = None,
         element: Optional = None
     ) -> List[WebElement]:
+        key=key.strip().replace(' ','.')
         return self.__find(
             by,
             EC.presence_of_all_elements_located,
@@ -819,11 +839,15 @@ class Firefox:
 
     # essentially send keys
     def js_setvalue_via_typing(self, element: WebElement, value) -> bool:
-        element.send_keys(Keys.COMMAND,"a")
+        is_mac=platform.system()=='Darwin'
+        if is_mac: 
+            element.send_keys(KEYS.COMMAND,"a")
+        else:
+            element.send_keys(KEYS.CONTROL,"a")
         time.sleep(0.5)
         element.send_keys(value)
     send_keys=js_setvalue_via_typing
-
+    send=js_setvalue_via_typing
 
     # for D2L, this methods would not save value!
     def js_setvalue(self, element: WebElement, value) -> bool:
@@ -995,7 +1019,64 @@ class Firefox:
             es = WebDriverWait(element, timeout).until(
                 find_func((by, key))
             )
-
+            # jerry: alias enhanced
+            es.send=es.send_keys
+            # https://stackoverflow.com/a/54662690/2292993
+            from functools import partial
+            def _senddelay(
+                self,
+                keys: str,
+                min_delay: float = 0.025,
+                max_delay: float = 0.25
+            ) -> None:
+                import random
+                for key in keys:
+                    self.send_keys(key)
+                    time.sleep(random.uniform(min_delay,max_delay))
+            def _sendsubmit(self,*value):
+                # send keys and then submit
+                self.senddelay(*value)
+                time.sleep(0.5)
+                self.submit()
+            @property
+            def _location(self):
+                # https://stackoverflow.com/a/59347207/2292993
+                # Assume there is equal amount of browser chrome on the left and right sides of the screen.
+                canvas_x_offset = self.parent.execute_script("return window.screenX + (window.outerWidth - window.innerWidth) / 2 - window.scrollX;")
+                # Assume all the browser chrome is on the top of the screen and none on the bottom.
+                canvas_y_offset = self.parent.execute_script("return window.screenY + (window.outerHeight - window.innerHeight) - window.scrollY;")
+                # Get the element center.
+                element_location = (self.rect["x"] + canvas_x_offset + self.rect["width"] / 2,
+                                    self.rect["y"] + canvas_y_offset + self.rect["height"] / 2)
+                return element_location
+            def _moveclick(self,n=2):
+                """
+                n = -1
+                    use selenium method
+                n could be 0; 2 in case the first click only gets the focus
+                    use python automation
+                """
+                if n==-1:
+                    # self is the element, self.parent is driver
+                    actions = ActionChains(self.parent)
+                    actions.move_to_element(self).click().perform()
+                else:
+                    # https://stackoverflow.com/a/59347207/2292993
+                    # Assume there is equal amount of browser chrome on the left and right sides of the screen.
+                    canvas_x_offset = self.parent.execute_script("return window.screenX + (window.outerWidth - window.innerWidth) / 2 - window.scrollX;")
+                    # Assume all the browser chrome is on the top of the screen and none on the bottom.
+                    canvas_y_offset = self.parent.execute_script("return window.screenY + (window.outerHeight - window.innerHeight) - window.scrollY;")
+                    # Get the element center.
+                    element_location = (self.rect["x"] + canvas_x_offset + self.rect["width"] / 2,
+                                        self.rect["y"] + canvas_y_offset + self.rect["height"] / 2)
+                    import ez
+                    ez.move(element_location)
+                    ez.click(n)
+            es.senddelay=partial(_senddelay, es)
+            es.sendsubmit=partial(_sendsubmit, es)
+            es.location=partial(_location, es)
+            es.loc=es.location
+            es.moveclick=partial(_moveclick, es)
             return es
         except:
             return None
